CLASS_IN,CLASS_ANY = 1,255
TYPE_A,TYPE_NS,TYPE_CNAME,TYPE_SOA,TYPE_PTR,TYPE_MX,TYPE_TXT,TYPE_AAAA,TYPE_SPF = 1,2,5,6,12,15,16,28,99

function math.eval(expr)
  local constant = function(v)
    if type(v) == 'number' then return v end
    v = v:lower()
    if v == 'pi' then return 3.14159265358979323846 end
    if v == 'e' then return 2.7182818284590452354 end
    return v
  end
  local operate = function(stack, out)
    local op = table.remove(stack)
    if op == '+' then
      local b,a = table.remove(out), table.remove(out)
      out[#out+1] = a + b
    elseif op == '-' then
      local b,a = table.remove(out), table.remove(out)
      out[#out+1] = a - b
    elseif op == '/' then
      local b,a = table.remove(out), table.remove(out)
      out[#out+1] = a / b
    elseif op == '%' then
      local b,a = table.remove(out), table.remove(out)
      out[#out+1] = a % b
    elseif op == '*' then
      local b,a = table.remove(out), table.remove(out)
      out[#out+1] = a * b
    elseif op == '^' then
      local b,a = table.remove(out), table.remove(out)
      out[#out+1] = math.pow(a, b)
    elseif op == 'atan2' or op == 'fmod' or op == 'frexp' or op == 'ldexp' or op == 'min' or op == 'max' or op == 'pow' or op == 'random' then
      local b,a = table.remove(out), table.remove(out)
      out[#out+1] = math[op](a, b)
    elseif math[op] then
      local b = table.remove(out)
      out[#out+1] = math[op](b)
    end
    return true
  end
  local out,stack,tokens,token,operator = {},{},{},'', { ['+']=1, ['-']=1, ['*']=2, ['/']=2, ['%']=2, ['^']=3 }
  for i=1,#expr do
    local c = expr:sub(i,i)
    if c == '-' then
      if token == '' then
        token = token .. '-'
      else
        if token ~= '' then tokens[#tokens+1],token = token,'' end
        tokens[#tokens+1] = c
      end
    elseif operator[c] or c == '(' or c == ')' or c == ',' then
      if token ~= '' then tokens[#tokens+1],token = token,'' end
      tokens[#tokens+1] = c
    elseif c ~= ' ' and c ~= '\n' and c ~= '\r' and c ~= '\t' then
      token = token .. c
    end
  end
  if token ~= '' then tokens[#tokens+1],token = token,'' end
  while #tokens  > 0 do
    token = table.remove(tokens, 1)
    token = constant(token)
    if tonumber(token) then
      out[#out+1] = token
    elseif operator[token] then
      while #stack > 0 and operator[stack[#stack]] and ((token ~= '^' and operator[stack[#stack]] >= operator[token]) or operator[stack[#stack]] > operator[token]) do
        if not operate(stack, out) then return nil end
      end
      stack[#stack+1] = token
    elseif math[token] then
      stack[#stack+1] = token
    elseif token == ',' then
      while #stack > 0 and stack[#stack] ~= '(' do if not operate(stack, out) then return nil end end
    elseif token == '(' then
      stack[#stack+1] = token
    elseif token == ')' then
      while #stack > 0 and stack[#stack] ~= '(' do if not operate(stack, out) then return nil end end
      if table.remove(stack) ~= '(' then return nil end
      if #stack > 0 and math[stack[#stack]] then if not operate(stack, out) then return nil end end
    end
  end
  while #stack > 0 do
    if not operate(stack, out) then return nil end
    if out[#out] == '(' then return nil end
  end
  if #out == 1 then return out[1] end
end

function answer(qname, qtype, qclass, lazy)
  ret = {}
  if qclass == CLASS_IN and qtype == TYPE_A then
    local domain = '.math'
    local a = qname:find(domain, 1, true)
    if a and a + #domain - 1 == #qname then
      local a = math.eval(qname:sub(1, a-1))
      if a then
        ret[#ret+1] = { name=qname, type=TYPE_CNAME, class=TYPE_IN, ttl=7200, cname=a..domain }
      end
    end
  end
  return ret
end
