/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/*
 LD_LIBRARY_PATH=. valgrind ./unnamed
 nslookup -type=mx -port=5353 eder.us 10.0.0.1
 nmap --script dns-fuzz -sU -p 53 10.0.0.1
*/

#include <signal.h>
#include "unnamed_internal.h"

#ifdef USE_PCAP
  #include <pcap.h>
#endif


#ifdef _MSC_VER
  #pragma comment(lib, "ws2_32.lib")
#else
  #include <libgen.h>
#endif

static void *cleanup(client_t *client)
{
  char ip_from[IPCHAR_SIZE], ip_to[IPCHAR_SIZE];
  state_t *state = client->state;
  int i;
  for(i = 0; i < client->state->module_count; i++)
  {
    if(client->state->modules[i].collect_garbage != NULL)
    {
      client->state->modules[i].collect_garbage(i, client->state);
    }
  }
  ip_from[0] = ip_to[0] = '\0';
  #ifdef USE_PTHREADS
    if(client->questions[0] != NULL)
    {
      INFO_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: reponse time: %f t=%d %s %s %s", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), microtime() - client->start, state->running_threads, (char*)client->questions[0]->qname.name.value, type_name(client->questions[0]->qtype), class_name(client->questions[0]->qclass));
    }
    else
    {
      INFO_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: reponse time: %f t=%d", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), microtime() - client->start, state->running_threads);
    }
  #else
    if(client->questions[0] != NULL)
    {
      INFO_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: reponse time: %f %s %s %s", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), microtime() - client->start, (char*)client->questions[0]->qname.name.value, type_name(client->questions[0]->qtype), class_name(client->questions[0]->qclass));
    }
    else
    {
      INFO_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: reponse time: %f", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), microtime() - client->start);
    }
  #endif

  free_client(client);

  (void)pthread_mutex_lock(&state->running_mutex);
  state->running_threads--;
  (void)pthread_mutex_unlock(&state->running_mutex);

  return NULL;
}

static void* dns_respond(void *void_client)
{
  ssize_t written;
  client_t *client = (client_t*)void_client;
  state_t *state = client->state;
  int i;

  if(client->socket != INVALID_SOCKET)
  {
    socklen_t slen = (socklen_t)sizeof(client->to_addr);
    (void)getsockname(client->socket, &client->to_addr.a.sockaddr, &slen);
  }

  if(client->is_tcp != 0) /* tcp */
  {
    int len, length;
    struct timeval tv;

    #ifdef S_SPLINT_S
      memset(&tv, 0, sizeof(tv));
    #endif
    tv.tv_sec = 1;
    tv.tv_usec = 0;
    (void)setsockopt(client->socket, SOL_SOCKET, SO_RCVTIMEO, (char*)&tv, (socklen_t)sizeof(struct timeval));

    while(client->packet.length < PACKET_LENGTH_SIZE)
    {
      if((len = (int)recv(client->socket, (char*)client->packet.value + client->packet.length, (buf_len)(client->packet.allocated - client->packet.length), 0)) <= 0)
      {
        return cleanup(client);
      }
      client->packet.length += (uint32_t)len;
    }

    length = (((int)(client->packet.value[0] << 8)) | (int)client->packet.value[1]) + PACKET_LENGTH_SIZE;
    if(buffer_resize(&client->packet, (uint32_t)length) != 0) return cleanup(client);
    while((int)client->packet.length < length)
    {
      if((len = (int)recv(client->socket, (char*)client->packet.value + client->packet.length, (buf_len)(client->packet.allocated - client->packet.length), 0)) <= 0)
      {
        return cleanup(client);
      }
      client->packet.length += (uint32_t)len;
    }
  }
  else /* udp */
  {
    client->socket = INVALID_SOCKET;
  }

  if(parse_packet(client) != 0) WARNING_IF(state->trace) (void)lprintf(TRACE, "Parse failed");
  if(client->header.flags.opcode != OPCODE_QUERY) WARNING_IF(state->trace) (void)lprintf(TRACE, "HEADER_OPCODE = %x", client->header.flags.opcode);
  client->header.flags.qr = 1;
  client->header.flags.ra = allow_forward(client) & 1;
  client->header.flags.ad = 0;
  client->header.flags.cd = 0;

  for(i = 0; i < (int)client->header.ancount; i++)
  {
    state->free_rr(&client->answers[i]);
    client->answers[i] = NULL;
  }
  client->header.ancount = 0;

  for(i = 0; i < (int)client->header.nscount; i++)
  {
    state->free_rr(&client->name_servers[i]);
    client->name_servers[i] = NULL;
  }
  client->header.nscount = 0;

  for(i = 0; i < (int)client->header.arcount; i++)
  {
    state->free_rr(&client->additional_records[i]);
    client->additional_records[i] = NULL;
  }
  client->header.arcount = 0;

  if(client->max_udp_size < 512)
  {
    client->max_udp_size = 512;
  }
  else if(client->max_udp_size > 512)
  {
#if 0
    rr_t *rr;
    WARNING_IF(state->trace) (void)lprintf(TRACE, "max_udp_size = %d", client->max_udp_size);
    rr = rr_create(NULL, TYPE_OPT, client->max_udp_size, 0, RCODE_NO_ERROR);
    if(rr != NULL)
    {
      (void)add_rr(client, RR_AR, &rr);
      free_rr(&rr);
    }
#endif
  }

  client->header.flags.rcode = RCODE_NAME_ERROR;
  for(i = 0; i < (int)client->header.qdcount; i++)
  {
    if(answer_question(client, RR_AN, &client->questions[i]->qname.name, client->questions[i]->qtype, client->questions[i]->qclass, 0) == ANSWER_SATISFIED)
    {
      client->header.flags.rcode = RCODE_NO_ERROR;
    }
  }
  if(client->header.qdcount > 0 || client->header.ancount > 0 || client->header.nscount > 0 || client->header.arcount > 0)
  {
    client->header.flags.rcode = RCODE_NO_ERROR;
  }

  if(write_packet(client) != 0)
  {
    WARNING_IF(state->trace) (void)lprintf(TRACE, "Failed to write packet to buffer");
    return cleanup(client);
  }

  if(client->is_tcp == 0)  /* udp */
  {
    int val;
    if(client->max_udp_size < (uint16_t)client->packet.length - PACKET_LENGTH_SIZE)
    {
      client->packet.length = (uint32_t)client->max_udp_size + PACKET_LENGTH_SIZE;
      client->packet.value[2] |= 2;  /* set TC flag */
    }
    if((client->socket = socket((int)client->to_addr.a.sockaddr.sa_family, SOCK_DGRAM, 0)) == INVALID_SOCKET)
    {
      char ip_from[IPCHAR_SIZE], ip_to[IPCHAR_SIZE];
      WARNING_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: failed to create return socket (sa_family=%d): %s", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), (int)client->to_addr.a.sockaddr.sa_family, ERRNO_STRING);
      return cleanup(client);
    }
    val = 1;
    (void)setsockopt(client->socket, SOL_SOCKET, SO_REUSEADDR, (const char*)&val, (socklen_t)sizeof(val));
    if(bind(client->socket, &client->to_addr.a.sockaddr, (socklen_t)sizeof(client->to_addr)) == -1)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "bind failed: %s", ERRNO_STRING);
      return cleanup(client);
    }
    if((written = sendto(client->socket, (char*)(client->packet.value + PACKET_LENGTH_SIZE), (buf_len)client->packet.length - PACKET_LENGTH_SIZE, 0, &client->from_addr.a.sockaddr, (socklen_t)sizeof(client->from_addr))) != (ssize_t)(client->packet.length - PACKET_LENGTH_SIZE))
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "sendto failed: %s", ERRNO_STRING);
      WARNING_IF(state->trace) (void)lprintf(TRACE, "Failed to write packet to socket %d, %d/%u written", (int)client->socket, (int)written, client->packet.length);
    }
  }
  else  /* tcp */
  {
    if((written = send(client->socket, (char*)client->packet.value, (buf_len)client->packet.length, 0)) != (ssize_t)client->packet.length)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "send failed: %s", ERRNO_STRING);
      WARNING_IF(state->trace) (void)lprintf(TRACE, "Failed to write packet to socket %d, %d/%u written", (int)client->socket, (int)written, client->packet.length);
    }
  }
  return cleanup(client);
}

static volatile sig_atomic_t quit_flag = 0;
static void set_quit_flag(int sig)
{
  #if !defined(WIN32) || defined(__CYGWIN__)
    (void)signal(sig, SIG_IGN);
    quit_flag = 1;
  #elif defined(WIN32)
    if(sig == SIGABRT)
    {
      quit_flag = 1;
    }
  #else
    quit_flag = 1;
  #endif
}

static void handle_packet(client_t *client)
{
  client->start = microtime();

  #ifdef USE_PTHREADS
  {
    state_t *state = client->state;
    pthread_t child = 0;
    pthread_attr_t attr;
    #ifdef S_SPLINT_S
      memset(&attr, 0, sizeof(attr));
    #endif
    (void)pthread_attr_init(&attr);
    (void)pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    (void)pthread_attr_setstacksize(&attr, 262144);

    (void)pthread_mutex_lock(&state->running_mutex);
    client->state->packet_count++;
    state->running_threads++;
    (void)pthread_mutex_unlock(&state->running_mutex);
    while(state->running_threads > 128)
    {
      ERROR_IF(client->state->trace) (void)lprintf(TRACE, "running_threads = %d", state->running_threads);
      (void)usleep(1000);
    }
    if(pthread_create(&child, &attr, dns_respond, client) != 0)
    {
      (void)pthread_mutex_lock(&state->running_mutex);
      state->running_threads--;
      (void)pthread_mutex_unlock(&state->running_mutex);
      free_client(client);
      ERROR_IF(client->state->trace) (void)lprintf(TRACE, "pthread_create failed: %s", ERRNO_STRING);
      quit_flag = 1;
    }
    (void)pthread_attr_destroy(&attr);
  }
  #else
  {
    client->state->packet_count++;
    (void)dns_respond(client);
  }
  #endif
}

static void init_tcp(state_t *state, SOCKET *tcp_sock)
{
  *tcp_sock = INVALID_SOCKET;
  if(quit_flag == 0)
  {
    if((*tcp_sock = socket((int)state->servaddr.a.sockaddr.sa_family, SOCK_STREAM, 0)) == INVALID_SOCKET) quit_flag = 1;
    if(*tcp_sock >= FD_SETSIZE) quit_flag = 1;
    if(quit_flag == 0)
    {
      struct timeval tv;
      int val = 1;
      (void)setsockopt(*tcp_sock, SOL_SOCKET, SO_REUSEADDR, (const char*)&val, (socklen_t)sizeof(val));
      #ifdef S_SPLINT_S
        memset(&tv, 0, sizeof(tv));
      #endif
      tv.tv_sec = 1;
      tv.tv_usec = 0;
      (void)setsockopt(*tcp_sock, SOL_SOCKET, SO_RCVTIMEO, (char*)&tv, (socklen_t)sizeof(struct timeval));
    }
  }
  if(quit_flag == 0)
  {
    if(bind(*tcp_sock, &state->servaddr.a.sockaddr, (socklen_t)sizeof(state->servaddr)) == -1)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "bind failed: %s", ERRNO_STRING);
      quit_flag = 1;
    }
  }
  if(quit_flag == 0)
  {
    if(listen(*tcp_sock, 10) == -1)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "listen failed: %s", ERRNO_STRING);
      quit_flag = 1;
    }
  }
}

#ifdef USE_PTHREADS
static void* tcp_loop(void *statep)
{
  state_t *state = (state_t*)statep;
  SOCKET tcp_sock = INVALID_SOCKET;
  client_t *client = NULL;

  init_tcp(state, &tcp_sock);

  while(quit_flag == 0)
  {
    socklen_t addr_len = (socklen_t)sizeof(client->from_addr);
    if(init_client(state, &client) != 0)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "init_client failed");
      free_client(client);
      break;
    }

    client->is_tcp = 1;
    client->socket = accept(tcp_sock, (struct sockaddr*)&client->from_addr, &addr_len);
    if(client->socket == INVALID_SOCKET)
    {
      if(errno != EAGAIN)
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "accept failed: %s", ERRNO_STRING);
      }
      continue;
    }

//broken
printf("handle_packet 1\n");
char ip_from[IPCHAR_SIZE], ip_to[IPCHAR_SIZE];
WARNING_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: failed to create return socket (sa_family=%d): %s", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), (int)client->to_addr.a.sockaddr.sa_family, ERRNO_STRING);

    handle_packet(client);
    client = NULL;
  }
  if(tcp_sock != INVALID_SOCKET) (void)closesocket(tcp_sock);

  free_client(client);

  (void)pthread_mutex_lock(&state->running_mutex);
  state->running_threads--;
  (void)pthread_mutex_unlock(&state->running_mutex);

  return NULL;
}
#endif

int main(int argc, char *argv[])
{
  int i;
  SOCKET udp_sock = INVALID_SOCKET;
  client_t *client = NULL;
  #ifndef USE_PTHREADS
    SOCKET tcp_sock = INVALID_SOCKET;
  #endif

  #ifdef USE_PCAP
    pcap_t *capture = NULL;
  #endif

  state_t _state;
  state_t *state = &_state;
  const char *config = "unnamed.config";

  #if !defined(WIN32) || defined(__CYGWIN__)
  {
    /* set up sig handler */
    struct sigaction sa;
    memset(&sa, 0, sizeof(sa));
    sa.sa_handler = set_quit_flag;
    (void)sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    if(sigaction(SIGINT, &sa, NULL) == -1)
    {
      (void)printf("sigaction failed: %s", ERRNO_STRING);
      return -1;
    }
  }
  #elif defined(WIN32)
  {
    (void)signal(SIGABRT, set_quit_flag);
  }
  #endif

  for(i = 1; i < argc; i++)
  {
    if(strcmp(argv[i], "-c") == 0 && i + 1 < argc)
    {
      config = argv[i+1];
      i++;
    }
    else if(strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0)
    {
      printf("%s [-c config_file] [option=value] [option=value] ...\n", argv[0]);
      return 0;
    }
  }

  /* attempt to change to unnamed's config directory */
  if(unlikely(chdir("/etc/unnamed/") != 0))
  {
    printf("chdir failed\n");
  }

  if(unnamed_init(state, config, argc, argv) != 0)
  {
    printf("Unable to initialize.\n");
    return -1;
  }

  #ifdef USE_PCAP
    char errbuf[PCAP_ERRBUF_SIZE];
    errbuf[0] = '\0';
    capture = pcap_open_live("any", 65536, 1, 500, errbuf);
    if(capture == NULL)
    {
      quit_flag = 1;
      ERROR_IF(state->trace) (void)lprintf(TRACE, "pcap_open failed: %s", errbuf);
    }
    else
    {
      struct bpf_program fp; 
      buffer_t filter;
      buffer_init(&filter);
      buffer_set(&filter, "udp port %d", ntohs(state->servaddr.a.sockaddr_in.sin_port));
      if(pcap_compile(capture, &fp, (char*)filter.value, 1, PCAP_NETMASK_UNKNOWN) == -1)
      {
        quit_flag = 1;
        ERROR_IF(state->trace) (void)lprintf(TRACE, "pcap_compile failed");
      }
      else
      {
        (void)pcap_setfilter(capture, &fp);
      }
      pcap_freecode(&fp);
      buffer_free(&filter);
    }
    (void)pcap_setdirection(capture, PCAP_D_IN);


  #else
    if((udp_sock = socket((int)state->servaddr.a.sockaddr.sa_family, SOCK_DGRAM, 0)) == INVALID_SOCKET) quit_flag = 1;
    if(udp_sock >= FD_SETSIZE) quit_flag = 1;
    if(quit_flag == 0)
    {
      struct timeval tv;
      int val = 1;
      (void)setsockopt(udp_sock, SOL_SOCKET, SO_REUSEADDR, (const char*)&val, (socklen_t)sizeof(val));
      val = 131072;
      (void)setsockopt(udp_sock, SOL_SOCKET, SO_RCVBUF, (const char*)&val, (socklen_t)sizeof(val));

      #ifdef S_SPLINT_S
        memset(&tv, 0, sizeof(tv));
      #endif
      tv.tv_sec = 1;
      tv.tv_usec = 0;
      (void)setsockopt(udp_sock, SOL_SOCKET, SO_RCVTIMEO, (char*)&tv, (socklen_t)sizeof(struct timeval));
    }
    if(quit_flag == 0)
    {
      if(bind(udp_sock, &state->servaddr.a.sockaddr, (socklen_t)sizeof(state->servaddr)) == -1)
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "bind failed: %s", ERRNO_STRING);
        quit_flag = 1;
      }
    }
  #endif

  #ifndef USE_PTHREADS
    init_tcp(state, &tcp_sock);
  #endif

  if(quit_flag == 0) (void)drop_priv(state);

  #ifdef USE_PTHREADS
  {
    pthread_t child = 0;
    pthread_attr_t attr;
    #ifdef S_SPLINT_S
      memset(&attr, 0, sizeof(attr));
    #endif
    (void)pthread_attr_init(&attr);
    (void)pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    (void)pthread_attr_setstacksize(&attr, 262144);

    (void)pthread_mutex_lock(&state->running_mutex);
    state->running_threads++;
    (void)pthread_mutex_unlock(&state->running_mutex);
    if(pthread_create(&child, &attr, tcp_loop, state) != 0)
    {
      (void)pthread_mutex_lock(&state->running_mutex);
      state->running_threads--;
      (void)pthread_mutex_unlock(&state->running_mutex);
      ERROR_IF(state->trace) (void)lprintf(TRACE, "pthread_create failed: %s", ERRNO_STRING);
    }
    (void)pthread_attr_destroy(&attr);
  }
  #endif

  while(quit_flag == 0)
  {
    int ret;
    #ifndef USE_PTHREADS
      int nfds;
      fd_set rfds;
    #endif

    #ifndef USE_PCAP
      socklen_t addr_len = (socklen_t)sizeof(client->from_addr);
    #endif
    if(init_client(state, &client) != 0)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "init_client failed");
      free_client(client);
      break;
    }

    #ifndef USE_PTHREADS
      FD_ZERO(&rfds);
      FD_SET((unsigned int)udp_sock, &rfds);
      nfds = udp_sock;
      if(tcp_sock != INVALID_SOCKET)
      {
        FD_SET((unsigned int)tcp_sock, &rfds);
        if(tcp_sock > nfds) nfds = tcp_sock;
      }
      if(select(nfds + 1, &rfds, NULL, NULL, NULL) == -1)
      {
        if(quit_flag == 0) ERROR_IF(state->trace) (void)lprintf(TRACE, "select failed: %s", ERRNO_STRING);
        continue;
      }
    #endif

    #ifndef USE_PTHREADS
    if(FD_ISSET((unsigned int)udp_sock, &rfds))
    {
    #endif

    #ifdef USE_PCAP
      struct pcap_pkthdr *header;
      const u_char *pkt_data = NULL;
      unsigned short port;
      int packet_length;
      uint32_t length;

/* todo: check header->caplen and set_raw */

      ret = pcap_next_ex(capture, &header, &pkt_data);
      if(ret == 0)
      {
        continue;
      }
      if(ret < 0)
      {
        quit_flag = 1;
        break;
      }
      packet_length = header->caplen;

      while(pkt_data[12] == 0x00 && pkt_data[13] == 0x00)
      {
        pkt_data += 2;
        packet_length -= 2;
      }

      if(pkt_data[12] != 0x08 || pkt_data[13] != 0x00) /* Ethernet type = 0x0800 */
      {
        continue;
      }
      pkt_data += 14;
      packet_length -= 14;

      if((pkt_data[0] & 0xf0) != 0x40)  /* IP version = 4 */
      {
        continue;
      }

      if(pkt_data[9] != IPPROTO_UDP)
      {
        continue;
      }

      client->from_addr.a.sockaddr_in.sin_family = AF_INET;
      memcpy(&client->from_addr.a.sockaddr_in.sin_addr, pkt_data + 12, sizeof(client->from_addr.a.sockaddr_in.sin_addr));

      client->to_addr.a.sockaddr_in.sin_family = AF_INET;
      memcpy(&client->to_addr.a.sockaddr_in.sin_addr, pkt_data + 16, sizeof(client->to_addr.a.sockaddr_in.sin_addr));

      packet_length -= (pkt_data[0] & 0x0f) << 2;
      pkt_data += (pkt_data[0] & 0x0f) << 2;

      memcpy(&port, pkt_data + 2, sizeof(client->from_addr.a.sockaddr_in.sin_port));
      if(port != state->servaddr.a.sockaddr_in.sin_port)  /* UDP port */
      {
        continue;
      }

      memcpy(&client->from_addr.a.sockaddr_in.sin_port, pkt_data, sizeof(client->from_addr.a.sockaddr_in.sin_port));
      memcpy(&client->to_addr.a.sockaddr_in.sin_port, pkt_data + 2, sizeof(client->to_addr.a.sockaddr_in.sin_port));

      length = (pkt_data[4] << 8) | pkt_data[5];

      pkt_data += 6;

      if(buffer_put_raw(&client->packet, pkt_data, length + 2) != 0)
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "buffer_put_raw failed.");
        quit_flag = 1;
        break;
      }

      client->packet.value[0] = (uint8_t)(((unsigned int)length & 0xff00) >> 8);
      client->packet.value[1] = (uint8_t)(length & 0xff);

    #else
      if((ret = (int)recvfrom(udp_sock, (char*)client->packet.value + PACKET_LENGTH_SIZE, (buf_len)(client->packet.allocated - PACKET_LENGTH_SIZE), 0, (struct sockaddr*)&client->from_addr, &addr_len)) < 12)
      {
        if(ret >= 0) ERROR_IF(state->trace) (void)lprintf(TRACE, "recvfrom failed (ret=%d): %s", ret, ERRNO_STRING);
        continue;
      }

      client->packet.length += (uint32_t)ret;

      client->packet.value[0] = (uint8_t)(((unsigned int)ret & 0xff00) >> 8);
      client->packet.value[1] = (uint8_t)(ret & 0xff);
      client->socket = udp_sock;
    #endif


//broken
printf("handle_packet 2\n");
char ip_from[IPCHAR_SIZE], ip_to[IPCHAR_SIZE];
WARNING_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: failed to create return socket (sa_family=%d): %s", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), (int)client->to_addr.a.sockaddr.sa_family, ERRNO_STRING);
    handle_packet(client);
    client = NULL;

    #ifndef USE_PTHREADS
    }
    else if(tcp_sock != INVALID_SOCKET && FD_ISSET((unsigned int)tcp_sock, &rfds))
    {
      client->is_tcp = 1;
      client->socket = accept(tcp_sock, (struct sockaddr*)&client->from_addr, &addr_len);
      if(client->socket == INVALID_SOCKET)
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "accept failed: %s", ERRNO_STRING);
        continue;
      }

//broken
printf("handle_packet 3\n");
char ip_from[IPCHAR_SIZE], ip_to[IPCHAR_SIZE];
WARNING_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: failed to create return socket (sa_family=%d): %s", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), (int)client->to_addr.a.sockaddr.sa_family, ERRNO_STRING);
      handle_packet(client);
      client = NULL;
    }
    #endif
  }

  DEBUG_IF(state->trace) (void)lprintf(TRACE, "cleanup ...");
  INFO_IF(state->trace) (void)lprintf(TRACE, "%d packets processed", state->packet_count);
  if(udp_sock != INVALID_SOCKET) (void)closesocket(udp_sock);

  free_client(client);

  #ifdef USE_PCAP
    if(capture != NULL)
    {
      pcap_close(capture);
      capture = NULL;
    }
  #endif

  #ifdef USE_PTHREADS
    for(i = 0; i < 100 && state->running_threads > 0; i++)
    {
      (void)usleep(250000);
    }
  #endif
  unnamed_term(state);
  config_free(state);
  #ifdef USE_PTHREADS
    pthread_exit(NULL);
  #endif

  return -1;
}
