/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "unnamed.h"

#define REALLOC_SIZE  32

void buffer_free(buffer_t *b);
int buffer_resize(buffer_t *b, uint32_t size);
int buffer_copy(buffer_t *dst, const buffer_t *src);
int buffer_set(buffer_t *b, const char *fmt, ...) __attribute__ ((format (printf, 2, 3)));
int buffer_append(buffer_t *b, const char *fmt, ...) __attribute__ ((format (printf, 2, 3)));
__attribute__((pure)) int buffer_cmp(const buffer_t *a, const buffer_t *b);
int buffer_get_raw(buffer_t *b, uint8_t *dst, uint32_t length);
int buffer_put_raw(buffer_t *b, const uint8_t *src, uint32_t length);
int buffer_get32(buffer_t *b, uint32_t *n);
int buffer_get16(buffer_t *b, uint16_t *n);
int buffer_put32(buffer_t *b, uint32_t value);
int buffer_put16(buffer_t *b, uint16_t value);
int buffer_get_addr(buffer_t *b, address_t *a);
int buffer_get_addr6(buffer_t *b, address_t *a);
int buffer_put_addr(buffer_t *b, const address_t *a);
int buffer_put_addr6(buffer_t *b, const address_t *a);
int buffer_put_buffer(buffer_t *data, const buffer_t *b);

int unnamed_init(state_t *state, const char *config, int argc, char *argv[]);
void unnamed_term(state_t *state);

const char* addr2char(char ip[IPCHAR_SIZE], const address_t *a);
int char2addr(const char *ip, in_port_t port, /*@out@*/ address_t *addr);
__attribute__((pure)) int addr_cmp(const address_t *a, const address_t *b);
__attribute__((pure)) int is_in_range(const address_t *address, const ip_range_t *range);
int addr_prefix(const buffer_t *name, /*@out@*/ address_t *a);
int addr6_prefix(const buffer_t *name, /*@out@*/ address_t *a);
__attribute__((pure)) int is_subdomain(const char *domain, const char *check);
int ptr_name(buffer_t *name, const address_t *a);
SOCKET unnamed_socket(int type);

void config_free(state_t *state);
void config_dump(state_t *state);
int config_read(const char *filename, config_node_t *root);
int config_set(state_t *state, config_node_t *root, const char *key, char *value);
void config_free_node(state_t *state, const char *name);
int config_maxn(state_t *state, const char *name);
const char *config_get_const(state_t *state, const char *name);
char* config_get(state_t *state, const char *name);
char *config_get_lower(state_t *state, const char *name);
int config_get_int(state_t *state, const char *name, int default_value);
double config_get_double(state_t *state, const char *name, double default_value);
int config_get_ip(state_t *state, const char *name, address_t *addr);
int config_get_ip_range(state_t *state, const char *name, ip_range_t *range);

int black_white_init(state_t *state);
void black_white_term(state_t *state);
__attribute__((pure)) int is_ip_black(state_t *state, const address_t *a);
int is_name_white(state_t *state, const buffer_t *name);
int is_name_black(state_t *state, const buffer_t *name);

question_t* question_create(const buffer_t *name, uint16_t type, uint16_t dclass);
void free_rr(rr_t **rr);
int add_rr(client_t *client, rr_section_t rr_section, rr_t **rr);
rr_t* rr_create(const buffer_t *name, uint16_t type, uint16_t dclass, uint32_t expires, uint8_t status);
int parse_packet(client_t *client);
int write_packet(client_t *client);
int init_client(state_t *state, client_t **client);
void free_client(client_t *client);
double microtime(void);
uint32_t current_serial(void);
__attribute__((const)) const char* type_name(uint16_t type);
__attribute__((const)) const char* class_name(uint16_t dclass);
__attribute__((const)) const char* answer_type_name(answer_t a);
__attribute__((const)) const char* rcode_name(uint8_t rcode);
answer_t answer_question(client_t *client, rr_section_t rr_section, const buffer_t *name, uint16_t type, uint16_t dclass, int lazy);
int drop_priv(state_t *state);
int unnamed_match(const char *pattern, const char *string);
int match_add(state_t *state, pattern_list_t *patterns, char *pattern);
int match_compile(state_t *state, char **list, int count, pattern_list_t *patterns);
void match_free(pattern_list_t *patterns);
int match_pattern_list(const pattern_list_t *patterns, const char *string);

uint16_t random_id(state_t *state);

int lprintf(state_t *state, const char *file, const char *function, int line, const char *fmt, ...) __attribute__ ((format (printf, 5, 6)));

void dump_append_client(buffer_t *out, const char *label, client_t *client, unsigned int indent);

__attribute__((pure)) unsigned int allow_forward(client_t *client);

#if defined(_lint) || __GNUC__ < 3
  #define likely(x)       (x)
  #define unlikely(x)     (x)
#else
  #define likely(x)       __builtin_expect((x),1)
  #define unlikely(x)     __builtin_expect((x),0)
#endif
